#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

class datetime():
    def __init__(self):
        pass
    
    @staticmethod
    def utcnow():
        return datetime()

    def strftime(self, format):
        return "2018-10-14 12:12:12"